/*
								+----------------------------------+
								|                                  |
								|    ***  convex.madness02  ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <vector>
#include <algorithm>
#include "../Vector.h"
#include "../MyVertex.h"
#include "../Polygon.h"

Vector2f v_mouse;
// mouse cursor

int n_width = 640, n_height = 480;
// window size

CPolygon2<TRefVertex<TMyVertex> > polygon, split_polygon;
std::vector<TMyVertex*> vertex_list;
// polygon and it's vertex list

TMyVertex t_plane_a(.2f, .3f, 0), t_plane_b(-.2f, -.3f, 0);
// plane vertices (added to the list as well)

TMyVertex *p_snap_vertex = 0;
// vertex we're holding with mouse

const float f_split_epsilon = .05f;


/*
 *	class CSnapVertex
 *		- simple function object for snapping vertices
 */
class CSnapVertex {
protected:
	float m_f_best_distance;
	TMyVertex *m_p_snap_vertex;
	Vector2f m_v_mouse;

public:
	inline CSnapVertex(float f_max_distance, Vector2f v_mouse)
		:m_f_best_distance(f_max_distance), m_p_snap_vertex(0), m_v_mouse(v_mouse)
	{}

	inline void operator ()(TMyVertex *p_vertex)
	{
		if((m_v_mouse - ((Vector3f)*p_vertex).v_xy()).f_Length() < m_f_best_distance) {
			m_f_best_distance = (m_v_mouse - ((Vector3f)*p_vertex).v_xy()).f_Length();
			m_p_snap_vertex = p_vertex;
		}
	}

	inline operator TMyVertex*() const
	{
		return m_p_snap_vertex;
	}
};

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// set ortho projection

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set camera

	glColor3f(1, 1, 1);
	// white

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	// wireframe

	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	// smooth points and lines
}

/*
 *	class CDrawPolygon
 *		- simple function object we can use to simply enter polygon's vertices to OpenGL
 */
class CDrawPolygon {
public:
	inline void operator ()(const TRefVertex<TMyVertex> &r_t_vertex) const
	{
		TMyVertex *p_ref = r_t_vertex.m_p_ref;
		glVertex3f(p_ref->v_position.x, p_ref->v_position.y, p_ref->v_position.z);
		// bare position is enough
	}
};

void RenderString(float f_x, float f_y,	const char *p_s_string, void *p_font = GLUT_BITMAP_8_BY_13)
{
	glRasterPos2f(f_x, f_y);
	for(; *p_s_string; ++ p_s_string)
    	glutBitmapCharacter(p_font, *p_s_string);
}

void onDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_BLEND);
	// smooth points and lines

	glLineWidth(4);
	glBegin(GL_POLYGON);
	polygon.for_each_vertex(CDrawPolygon());
	glEnd();
	// draw polygon

	glColor3f(1, 0, 0);
	// red

	glPointSize(16);
	glBegin(GL_POINTS);
	polygon.for_each_vertex(CDrawPolygon());
	glEnd();
	// draw polygon's vertices

	glColor3f(1, 1, 1);
	// white

	Vector3f v_a = t_plane_a;
	Vector3f v_b = t_plane_b;
	v_b -= v_a;
	v_b.Normalize();
	Vector3f v_off = Vector3f(v_b.y, -v_b.x, 0) * f_split_epsilon;
	v_a -= v_b * 5.0f;
	v_b = v_a + v_b * 10.0f;
	glLineWidth(1.5f);
	glBegin(GL_LINES);
	glVertex3f(v_a.x, v_a.y, 0);
	glVertex3f(v_b.x, v_b.y, 0);

	glColor3f(.5f, .5f, .5f);
	// gray

	v_a += v_off;
	v_b += v_off;
	glVertex3f(v_a.x, v_a.y, 0);
	glVertex3f(v_b.x, v_b.y, 0);
	v_a -= v_off * 2.0f;
	v_b -= v_off * 2.0f;
	glVertex3f(v_a.x, v_a.y, 0);
	glVertex3f(v_b.x, v_b.y, 0);
	glEnd();
	// draw "infinite" plane line, gray collinear lines mark epsilon boundary

	glColor3f(0, 0, 1);
	// blue

	glPointSize(8);
	glBegin(GL_POINTS);
	split_polygon.for_each_vertex(CDrawPolygon());
	glEnd();
	// draw cut-polygon's vertices

	glBegin(GL_POLYGON);
	split_polygon.for_each_vertex(CDrawPolygon());
	glEnd();
	// draw cut-polygon's outline

	glColor3f(0, 1, 0);
	// green

	glBegin(GL_POINTS);
	glVertex3f(t_plane_a.v_position.x, t_plane_a.v_position.y, t_plane_a.v_position.z);
	glVertex3f(t_plane_b.v_position.x, t_plane_b.v_position.y, t_plane_b.v_position.z);
	glEnd();
	// draw subdivision plane points

	glDisable(GL_BLEND);

	glColor3f(1, 1, 1);
	// white

	RenderString(-1, .95f, (polygon.b_IsConvex())? "convex" : "non-convex");
	RenderString(-1, .90f, (split_polygon.b_IsConvex())? "cut convex" : "cut non-convex");
	// is it convex?

	char p_s_vertstring[32];
	sprintf(p_s_vertstring, "cut has %d vertices", split_polygon.n_Vertex_Num());
	RenderString(-1, .85f, p_s_vertstring);
	// how many vertices does it have?

	RenderString(-1, .80f, (split_polygon.b_IsTiny())? "cut is tiny" : "cut is not tiny");
	// is it tiny?

	glutSwapBuffers();
}

void onMouseClick(int button, int state, int x, int y)
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;
	// get current mouse position

	switch(button) {
    case GLUT_LEFT_BUTTON:
		p_snap_vertex = std::for_each(vertex_list.begin(),
			vertex_list.end(), CSnapVertex(.15f, v_mouse));
		// try to snap some vertex near the mouse
		break;
    case GLUT_RIGHT_BUTTON:
    case GLUT_MIDDLE_BUTTON:
		p_snap_vertex = 0; // nothing with those buttons
		break;
	};
}

void onMouseMotion(int x, int y) // dragging
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;

	if(p_snap_vertex) {
		*p_snap_vertex = Vector3f(v_mouse.x, v_mouse.y, 0);

		split_polygon = polygon;
		split_polygon.Cut(Plane3f((Vector3f)t_plane_a - (Vector3f)t_plane_b,
			Vector3f(0, 0, 1), (Vector3f)t_plane_a), plane_Front, f_split_epsilon);
		// refresh cut-polygon

		glutPostRedisplay();
	}
	// move our vertex
}

int main(int n_arg_num, /*const*/ char **p_arg_list) // glut doesn't like 'const'
{
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex(-.5f, -.5f, 0), vertex_list));
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex(-.5f,  .5f, 0), vertex_list));
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex( .5f,  .5f, 0), vertex_list));
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex( .5f, -.5f, 0), vertex_list));
	polygon.Calc_Normal();
	// add vertices

	vertex_list.push_back(&t_plane_a);
	vertex_list.push_back(&t_plane_b);
	// add plane vertices so it's possible to move them
	// without rewriting anything, i'm lazy after all

	split_polygon = polygon;
	split_polygon.Cut(Plane3f((Vector3f)t_plane_a - (Vector3f)t_plane_b,
		Vector3f(0, 0, 1), (Vector3f)t_plane_a), plane_Front, f_split_epsilon);
	// create cut-polygon

	glutInit(&n_arg_num, p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("convex.madness");
	// init OpenGL using GLUT

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutMouseFunc(onMouseClick);
	glutMotionFunc(onMouseMotion);
	// register callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
