/*
								+----------------------------------+
								|                                  |
								|    ***  Vector math v2.0  ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2003   |
								|                                  |
								|            Vector.cpp            |
								|                                  |
								+----------------------------------+
*/

/*
 *	passed code revision at 2006-05-16
 *
 *	optimized some parts of code
 *	all integers storing just true of false were replaced with bool
 *
 *	passed code revision at 2006-07-02
 *
 *	added swizzle operators to vector classes
 *	added redlect / refract / fresnel functions to Vector3<T> class
 *	renamed Vector3<T>::Len to Vector3<T>::f_Length, Vector3<T>::Len2 to Vector3<T>::f_Length2
 *	added some Matrix3f methods, renamed Matrix3f::Init() to Matrix3f::Identity()
 *	renamed Planef to Plane3f
 *	renamed Pi to f_pi
 *	general cleanup of code
 *
 *	2006-07-17
 *
 *	renamed CPolygon::r_Vertex(int) to CPolygon::r_t_Vertex(int)
 *	renamed CPolygon::r_Normal() to CPolygon::r_t_Normal()
 *
 *	2006-07-20
 *
 *	added CPolygon::b_MT_RayHit_CullFrontfaces and CPolygon::b_MT_RayHit_CullBackFaces
 *
 *	2006-07-24
 *
 *	renamed TVertexStruct to TVertex3f
 *	added class CVectorMath with ray distances and bounding box / line (ray) intersection tests
 *
 *	2007-01-23
 *
 *	fixed std::vector handling in template CPolygon<TVertex3f> (not-enough-memory checks)
 *	remade TVertex3f base class and CPolygon<TVertex3f> a bit (added + and * operators,
 *	renamed Lerp() to t_Lerp())
 *
 *	note this was modified for convex madness series and most of stuff was teared down
 *
 */

#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <vector>
#include "../Vector.h"

const float f_pi = 3.1415926535897932384626433832795028841971691075f;
const float f_epsilon = .05f; // modified to see it better

/*
 *        === Plane3f ===
 */

Plane3f::Plane3f(const Vector3f &r_v_pos, const Vector3f &r_v_norm)
{
	v_normal = r_v_norm;
	//assert(v_normal.f_Length() > f_epsilon);
	f_dist = -r_v_pos.f_Dot(v_normal);
}
	
Plane3f::Plane3f(const Vector3f &r_v_u, const Vector3f &r_v_v, const Vector3f &r_v_pos)
{
	v_normal = r_v_u.v_Cross(r_v_v);
	//assert(v_normal.f_Length() > f_epsilon);
	v_normal.Normalize();
	f_dist = -r_v_pos.f_Dot(v_normal);
}

EPlanePos Plane3f::n_Vector_Pos(const Vector3f &r_v_vec, float f_epsilon_ex) const
{
	float f;
	return ((f = f_Vector_Dist(r_v_vec)) > f_epsilon_ex)? plane_Front : (f < -f_epsilon_ex)? plane_Back : plane_Onplane;
}

EPlanePos Plane3f::n_Tri_Pos(const Vector3f &v_a,
	const Vector3f &v_b, const Vector3f &v_c, float f_epsilon_ex) const
{
	int b_front = false, b_back = false;

	switch(n_Vector_Pos(v_a, f_epsilon_ex)) {
	case plane_Front:
		b_front = true;
		break;
	case plane_Back:
		b_back = true;
		break;
	}
	switch(n_Vector_Pos(v_b, f_epsilon_ex)) {
	case plane_Front:
		b_front = true;
		break;
	case plane_Back:
		b_back = true;
		break;
	}
	switch(n_Vector_Pos(v_c, f_epsilon_ex)) {
	case plane_Front:
		if(!b_back)
			return plane_Front;
		return plane_Split;
	case plane_Back:
		if(!b_front)
			return plane_Back;
		return plane_Split;
	}

	if(!b_front) {
		if(!b_back)
			return plane_Onplane;
		return plane_Back;
	} else if(!b_back)
		return plane_Front;
	return plane_Split;
}

float Plane3f::f_Vector_Dist(const Vector3f &r_v_vec) const
{
	return v_normal.x * r_v_vec.x + v_normal.y * r_v_vec.y + v_normal.z * r_v_vec.z + f_dist;
}

float Plane3f::f_Vector_Dot(const Vector3f &r_v_vec) const
{
	return v_normal.f_Dot(r_v_vec);
}

Vector3f Plane3f::v_Intersect_Ray(const Vector3f &v_org, const Vector3f &v_dir) const
{
	float f_t;

	if((f_t = f_Vector_Dot(v_dir)) == .0f) {
		//assert(0);
		throw CRayIsParallel_Exception();
		return v_org;
	}

	f_t = f_Vector_Dist(v_org) / f_t;
	// vzdlenost

	return v_org - v_dir * f_t;
}

float Plane3f::f_Intersect_Ray_t(const Vector3f &v_org, const Vector3f &v_dir) const
{
	float f_t;

	if((f_t = f_Vector_Dot(v_dir)) == .0f) {
		//assert(0);
		throw CRayIsParallel_Exception();
		return 0;
	}

	return -f_Vector_Dist(v_org) / f_t;
	// vzdlenost
}

bool Plane3f::operator ==(const Plane3f &t_plane) const
{
	if(fabs(fabs(f_Vector_Dot(t_plane.v_normal)) - 1.0f) < f_epsilon &&
	   fabs(f_dist - t_plane.f_dist) < f_epsilon)
		return true;
	return false;
}

/*
 *        === ~Plane3f ===
 */

/*
 *        === Matrix4f ===
 */

void Matrix4f::Identity()
{
	for(int i = 0; i < 4; ++ i) {		
		for(int j = 0; j < 4; ++ j)
			f[i][j] = (float)(i == j);
	}
}

void Matrix4f::Scale(float f_scale)
{
	for(int i = 0; i < 3; ++ i) {		
		for(int j = 0; j < 4; ++ j)
			f[i][j] *= f_scale;
	}
}

void Matrix4f::Scale(const Vector3f &r_v_scale)
{
	for(int i = 0; i < 3; ++ i) {		
		for(int j = 0; j < 4; ++ j)
			f[i][j] *= r_v_scale[i];
	}
}


void Matrix4f::Translate(const Vector3f &r_v_dir)
{
	Matrix4f t_tmp;

    t_tmp.Identity();

    t_tmp.f[3][0] = r_v_dir.x;
    t_tmp.f[3][1] = r_v_dir.y;
    t_tmp.f[3][2] = r_v_dir.z;

	*this = (*this) * t_tmp;
}

void Matrix4f::RotateX(float f_angle)
{
	Matrix4f t_tmp;

    t_tmp.Identity();

    t_tmp.f[2][1] = (float)-sin(f_angle);
    t_tmp.f[2][2] = (float)cos(f_angle);
    t_tmp.f[1][1] = (float)cos(f_angle);
    t_tmp.f[1][2] = (float)sin(f_angle);

	*this = (*this) * t_tmp;
}

void Matrix4f::RotateY(float f_angle)
{
	Matrix4f t_tmp;

    t_tmp.Identity();

    t_tmp.f[0][0] = (float)cos(f_angle);
    t_tmp.f[0][2] = (float)-sin(f_angle);
    t_tmp.f[2][2] = (float)cos(f_angle);
    t_tmp.f[2][0] = (float)sin(f_angle);

	*this = (*this) * t_tmp;
}

void Matrix4f::RotateZ(float f_angle)
{
	Matrix4f t_tmp;

    t_tmp.Identity();

    t_tmp.f[0][0] = (float)cos(f_angle);
    t_tmp.f[0][1] = (float)sin(f_angle);
    t_tmp.f[1][1] = (float)cos(f_angle);
    t_tmp.f[1][0] = (float)-sin(f_angle);

	*this = (*this) * t_tmp;
}

void Matrix4f::Rotate(float f_angle, const Vector3f &r_v_axis)
{
	Matrix4f t_tmp;

    t_tmp.Identity();

	// R = uuT + cos(f_angle) * (I - uuT) + sin(f_angle)S =

	for(int i = 0; i < 3; ++ i) {
        for(int j = 0; j < 3; ++ j) {
            t_tmp.f[i][j] = r_v_axis[i] * r_v_axis[j] *
				(1 - (float)cos(f_angle)) + t_tmp.f[i][j] * (float)cos(f_angle);
		}
    }
	// R = uu^T * (1 - cos(f_angle)) + cos(f_angle) * I + ...

	t_tmp.f[1][0] -= (float)sin(f_angle) * r_v_axis.z;
	t_tmp.f[0][1] += (float)sin(f_angle) * r_v_axis.z;
	t_tmp.f[2][0] += (float)sin(f_angle) * r_v_axis.y;
	t_tmp.f[0][2] -= (float)sin(f_angle) * r_v_axis.y;
	t_tmp.f[2][1] -= (float)sin(f_angle) * r_v_axis.x;
	t_tmp.f[1][2] += (float)sin(f_angle) * r_v_axis.x;
	// ...  + sin(f_angle)S

	*this = (*this) * t_tmp;
}

Matrix4f Matrix4f::operator * (const Matrix4f &r_t_mat) const
{
	Matrix4f t_tmp;

    for(int i = 0; i < 4; ++ i) {
        for(int j = 0; j < 4; ++ j) {
            t_tmp.f[i][j] = f[0][j] * r_t_mat.f[i][0] + f[1][j] * r_t_mat.f[i][1] +
						  f[2][j] * r_t_mat.f[i][2] + f[3][j] * r_t_mat.f[i][3];
		}
    }

    return t_tmp;
}

Matrix4f Matrix4f::operator *=(const Matrix4f &r_t_mat)
{
	*this = (*this) * r_t_mat;
	return *this;
}

Matrix4f Matrix4f::t_Transpose() const
{
	Matrix4f t_tmp;

    for(int i = 0; i < 4; ++ i) {
        for(int j = 0; j < 4; ++ j)
            t_tmp.f[i][j] = f[j][i];
    }

    return t_tmp;
}

Matrix4f Matrix4f::t_Inverse() const
{
    float f_det = f[0][0] * f[1][1] * f[2][2] +
            f[1][0] * f[2][1] * f[0][2] +
            f[0][1] * f[1][2] * f[2][0] -
            f[0][2] * f[1][1] * f[2][0] -
            f[0][1] * f[1][0] * f[2][2] -
            f[0][0] * f[1][2] * f[2][1];

	Matrix4f t_inverse;
    t_inverse.f[0][0] = f[1][1] * f[2][2] - f[1][2] * f[2][1];
    t_inverse.f[1][0] =-f[1][0] * f[2][2] + f[1][2] * f[2][0];
    t_inverse.f[2][0] = f[1][0] * f[2][1] - f[1][1] * f[2][0];
    t_inverse.f[3][0] =-f[1][0] * f[2][1] * f[3][2] -
		f[1][1] * f[2][2] * f[3][0] -
		f[2][0] * f[3][1] * f[1][2] +
		f[1][2] * f[2][1] * f[3][0] +
		f[1][1] * f[2][0] * f[3][2] +
		f[2][2] * f[1][0] * f[3][1];
    t_inverse.f[0][1] =-f[0][1] * f[2][2] + f[0][2] * f[2][1];
    t_inverse.f[1][1] = f[0][0] * f[2][2] - f[0][2] * f[2][0];
    t_inverse.f[2][1] =-f[0][0] * f[2][1] + f[0][1] * f[2][0];
	t_inverse.f[3][1] = f[0][0] * f[2][1] * f[3][2] +
		f[0][1] * f[2][2] * f[3][0] +
		f[2][0] * f[3][1] * f[0][2] -
		f[0][2] * f[2][1] * f[3][0] -
		f[0][1] * f[2][0] * f[3][2] -
		f[2][2] * f[0][0] * f[3][1];
    t_inverse.f[0][2] = f[0][1] * f[1][2] - f[0][2] * f[1][1];
    t_inverse.f[1][2] =-f[0][0] * f[1][2] + f[0][2] * f[1][0];
    t_inverse.f[2][2] = f[0][0] * f[1][1] - f[0][1] * f[1][0];
    t_inverse.f[3][2] =-f[0][0] * f[1][1] * f[3][2] -
		f[0][1] * f[1][2] * f[3][0] -
		f[1][0] * f[3][1] * f[0][2] +
		f[0][2] * f[1][1] * f[3][0] +
		f[0][1] * f[1][0] * f[3][2] +
		f[1][2] * f[0][0] * f[3][1];

	f_det = 1.0f / f_det;
    for(int j = 0; j < 4; ++ j) {
        for(int i = 0; i < 3; ++ i)
            t_inverse.f[j][i] *= f_det;
	}

    t_inverse.f[0][3] = 0;
    t_inverse.f[1][3] = 0;
    t_inverse.f[2][3] = 0;
    t_inverse.f[3][3] = 1;

	return t_inverse;
}

void Matrix4f::Invert()
{
	*this = t_Inverse();
}

Vector3f Matrix4f::v_Transform_Pos(const Vector3f &r_v_vec) const
{
	return Vector3f(r_v_vec.x * f[0][0] + r_v_vec.y * f[1][0] + r_v_vec.z * f[2][0] + f[3][0],
				    r_v_vec.x * f[0][1] + r_v_vec.y * f[1][1] + r_v_vec.z * f[2][1] + f[3][1],
				    r_v_vec.x * f[0][2] + r_v_vec.y * f[1][2] + r_v_vec.z * f[2][2] + f[3][2]);
}

Vector3f Matrix4f::v_Transform_Dir(const Vector3f &r_v_vec) const
{
	return Vector3f(r_v_vec.x * f[0][0] + r_v_vec.y * f[1][0] + r_v_vec.z * f[2][0],
				    r_v_vec.x * f[0][1] + r_v_vec.y * f[1][1] + r_v_vec.z * f[2][1],
				    r_v_vec.x * f[0][2] + r_v_vec.y * f[1][2] + r_v_vec.z * f[2][2]);
}

/*
 *        === ~Matrix4f ===
 */
