/*
								+----------------------------------+
								|                                  |
								|    ***  convex.madness02  ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <vector>
#include <algorithm>
#include "../Vector.h"
#include "../MyVertex.h"
#include "../Polygon.h"

Vector2f v_mouse;
// mouse cursor

int n_width = 640, n_height = 480;
// window size

CPolygon2<TRefVertex<TMyVertex> > polygon;
std::vector<TMyVertex*> vertex_list;
// polygon and it's vertex list

TMyVertex *p_snap_vertex = 0;
// vertex we're holding with mouse

/*
 *	class CSnapVertex
 *		- simple function object for snapping vertices
 */
class CSnapVertex {
protected:
	float m_f_best_distance;
	TMyVertex *m_p_snap_vertex;
	Vector2f m_v_mouse;

public:
	inline CSnapVertex(float f_max_distance, Vector2f v_mouse)
		:m_f_best_distance(f_max_distance), m_p_snap_vertex(0), m_v_mouse(v_mouse)
	{}

	inline void operator ()(TMyVertex *p_vertex)
	{
		if((m_v_mouse - ((Vector3f)*p_vertex).v_xy()).f_Length() < m_f_best_distance) {
			m_f_best_distance = (m_v_mouse - ((Vector3f)*p_vertex).v_xy()).f_Length();
			m_p_snap_vertex = p_vertex;
		}
	}

	inline operator TMyVertex*() const
	{
		return m_p_snap_vertex;
	}
};

/*
 *	class CDrawPolygon
 *		- simple function object we can use to simply enter polygon's vertices to OpenGL
 */
class CDrawPolygon {
public:
	inline void operator ()(const TRefVertex<TMyVertex> &r_t_vertex) const
	{
		TMyVertex *p_ref = r_t_vertex.m_p_ref;
		glVertex3f(p_ref->v_position.x, p_ref->v_position.y, p_ref->v_position.z);
		// bare position is enough
	}
};

void RenderString(float f_x, float f_y,	const char *p_s_string, void *p_font = GLUT_BITMAP_8_BY_13)
{
	glRasterPos2f(f_x, f_y);
	for(; *p_s_string; ++ p_s_string)
    	glutBitmapCharacter(p_font, *p_s_string);
}

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// set ortho projection

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set camera

	glColor3f(1, 1, 1);
	// white

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	// wireframe

	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	// smooth points and lines

	glPointSize(8);
	// point size 8
}

void onDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_BLEND);
	// smooth points and lines

	glBegin(GL_POLYGON);
	polygon.for_each_vertex(CDrawPolygon());
	glEnd();
	// draw polygon

	glColor3f(1, 0, 0);
	// red

	glBegin(GL_POINTS);
	polygon.for_each_vertex(CDrawPolygon());
	glEnd();
	// draw polygon's vertices

	glColor3f(0, 1, 0);
	// yellow

	Vector3f v_nearest = polygon.v_NearestPoint_Barycentric(Vector3f(v_mouse.x, v_mouse.y, 10));
	glBegin(GL_POINTS);
	glVertex3f(v_nearest.x, v_nearest.y, v_nearest.z);
	glEnd();
	// draw nearest point

	glColor3f(1, 1, 1);
	// white

	glDisable(GL_BLEND);

	RenderString(-1, .95f, (polygon.b_IsConvex())? "convex" : "non-convex");
	// is it convex?

	glutSwapBuffers();
}

void onMouseClick(int button, int state, int x, int y)
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;
	// get current mouse position

	switch(button) {
    case GLUT_LEFT_BUTTON:
		p_snap_vertex = std::for_each(vertex_list.begin(),
			vertex_list.end(), CSnapVertex(.05f, v_mouse));
		// try to snap some vertex near the mouse
		break;
    case GLUT_RIGHT_BUTTON:
    case GLUT_MIDDLE_BUTTON:
		p_snap_vertex = 0; // nothing with those buttons
		break;
	};

	glutPostRedisplay();
}

void onMouseMotion(int x, int y) // dragging
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;

	if(p_snap_vertex)
		*p_snap_vertex = Vector3f(v_mouse.x, v_mouse.y, 0);
	// move our vertex

	glutPostRedisplay();
}

void onPassiveMouseMotion(int x, int y) // just moving mouse
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;

	glutPostRedisplay();
}

int main(int n_arg_num, /*const*/ char **p_arg_list) // glut doesn't like 'const'
{
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex(-.5f, -.5f, 0), vertex_list));
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex(-.5f,  .5f, 0), vertex_list));
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex( .5f,  .5f, 0), vertex_list));
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex( .5f, -.5f, 0), vertex_list));
	// add vertices

	polygon.Calc_Normal();

	glutInit(&n_arg_num, p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("convex.madness");
	// init OpenGL using GLUT

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutMouseFunc(onMouseClick);
	glutMotionFunc(onMouseMotion);
	glutPassiveMotionFunc(onPassiveMouseMotion);
	// register mouse callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
