/*
								+----------------------------------+
								|                                  |
								|    ***  convex.madness02  ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <vector>
#include <algorithm>
#include "../Vector.h"
#include "../MyVertex.h"
#include "../Polygon.h"

Vector2f v_mouse;
// mouse cursor

int n_width = 640, n_height = 480;
// window size

CPolygon2<TRefVertex<TMyVertex> > polygon, polygon2;
std::vector<TMyVertex*> vertex_list;
// polygon and it's vertex list

TMyVertex *p_center1, *p_center2;
// polygon centers (to move polygons)

TMyVertex *p_snap_vertex = 0;
// vertex we're holding with mouse

/*
 *	class CSnapVertex
 *		- simple function object for snapping vertices
 */
class CSnapVertex {
protected:
	float m_f_best_distance;
	TMyVertex *m_p_snap_vertex;
	Vector2f m_v_mouse;

public:
	inline CSnapVertex(float f_max_distance, Vector2f v_mouse)
		:m_f_best_distance(f_max_distance), m_p_snap_vertex(0), m_v_mouse(v_mouse)
	{}

	inline void operator ()(TMyVertex *p_vertex)
	{
		if((m_v_mouse - ((Vector3f)*p_vertex).v_xy()).f_Length() < m_f_best_distance) {
			m_f_best_distance = (m_v_mouse - ((Vector3f)*p_vertex).v_xy()).f_Length();
			m_p_snap_vertex = p_vertex;
		}
	}

	inline operator TMyVertex*() const
	{
		return m_p_snap_vertex;
	}
};

/*
 *	class CDrawPolygon
 *		- simple function object we can use to simply enter polygon's vertices to OpenGL
 */
class CDrawPolygon {
public:
	inline void operator ()(const TRefVertex<TMyVertex> &r_t_vertex) const
	{
		TMyVertex *p_ref = r_t_vertex.m_p_ref;
		glVertex3f(p_ref->v_position.x, p_ref->v_position.y, p_ref->v_position.z);
		// bare position is enough
	}
};

void RenderString(float f_x, float f_y,	const char *p_s_string, void *p_font = GLUT_BITMAP_8_BY_13)
{
	glRasterPos2f(f_x, f_y);
	for(; *p_s_string; ++ p_s_string)
    	glutBitmapCharacter(p_font, *p_s_string);
}

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// set ortho projection

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set camera

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	// wireframe

	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	// smooth points and lines

	glPointSize(8);
	// point size 8
}

void onDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glColor3f(1, 1, 1);
	// white

	glEnable(GL_BLEND);
	// smooth points and lines

	glBegin(GL_POLYGON);
	polygon.for_each_vertex(CDrawPolygon());
	glEnd();
	// draw polygon

	glBegin(GL_POLYGON);
	polygon2.for_each_vertex(CDrawPolygon());
	glEnd();
	// draw polygon2

	glColor3f(1, 0, 0);
	// red

	glBegin(GL_POINTS);
	polygon.for_each_vertex(CDrawPolygon());
	polygon2.for_each_vertex(CDrawPolygon());
	glEnd();
	// draw polygon's vertices
	
	glColor3f(0, 1, 0);
	// green

	glBegin(GL_POINTS);
	glVertex3f(p_center1->v_position.x, p_center1->v_position.y, p_center1->v_position.z);
	glVertex3f(p_center2->v_position.x, p_center2->v_position.y, p_center2->v_position.z);
	glEnd();

	glColor3f(1, 0, 0);
	// red

	RenderString(-1, .95f, (polygon.b_IsConvex() && polygon2.b_IsConvex())? "" :
		"one of polys is non-convex!");
	// is it convex?

	glColor3f(1, 1, 1);
	// white

	glDisable(GL_BLEND);

	RenderString(-1, .90f, (polygon.b_Overlap_SepAxes(polygon2))?
		"overlap" : "do not overlap");

	glutSwapBuffers();
}

void onMouseClick(int button, int state, int x, int y)
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;
	// get current mouse position

	switch(button) {
    case GLUT_LEFT_BUTTON:
		p_snap_vertex = std::for_each(vertex_list.begin(),
			vertex_list.end(), CSnapVertex(.05f, v_mouse));
		// try to snap some vertex near the mouse
		break;
    case GLUT_RIGHT_BUTTON:
    case GLUT_MIDDLE_BUTTON:
		p_snap_vertex = 0; // nothing with those buttons
		break;
	};
}

class CShiftVertices {
protected:
	Vector3f m_v_offset;

public:
	inline CShiftVertices(Vector3f v_offset)
		:m_v_offset(v_offset)
	{}

	inline void operator ()(TRefVertex<TMyVertex> &r_vertex) const
	{
		r_vertex += m_v_offset;
	}
};

void onMouseMotion(int x, int y) // dragging
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;

	if(p_snap_vertex) {
		if(p_snap_vertex == p_center1) {
			Vector3f v_offset = Vector3f(v_mouse.x, v_mouse.y, 0) - (Vector3f)*p_snap_vertex;
			polygon.for_each_vertex(CShiftVertices(v_offset));
		} else if(p_snap_vertex == p_center2) {
			Vector3f v_offset = Vector3f(v_mouse.x, v_mouse.y, 0) - (Vector3f)*p_snap_vertex;
			polygon2.for_each_vertex(CShiftVertices(v_offset));
		}
		*p_snap_vertex = Vector3f(v_mouse.x, v_mouse.y, 0);

		*p_center1 = polygon.v_Center();
		*p_center2 = polygon2.v_Center();

		glutPostRedisplay();
	}
	// move our vertex
}

int main(int n_arg_num, /*const*/ char **p_arg_list) // glut doesn't like 'const'
{
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex(-.75f, -.5f, 0), vertex_list));
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex(-.75f,  .5f, 0), vertex_list));
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex( .25f,  .5f, 0), vertex_list));
	polygon.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex( .25f, -.5f, 0), vertex_list));
	polygon.Calc_Normal();
	// add vertices

	polygon2.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex(-.25f, -.25f, 0), vertex_list));
	polygon2.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex(-.25f,  .25f, 0), vertex_list));
	polygon2.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex( .75f,  .25f, 0), vertex_list));
	polygon2.Insert_Vertex(TRefVertex<TMyVertex>(TMyVertex( .75f, -.25f, 0), vertex_list));
	polygon2.Calc_Normal();
	// add vertices

	TRefVertex<TMyVertex> center1(polygon.t_Center(), vertex_list);
	TRefVertex<TMyVertex> center2(polygon2.t_Center(), vertex_list);
	p_center1 = center1.m_p_ref;
	p_center2 = center2.m_p_ref;
	// create center vertices

	glutInit(&n_arg_num, p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("convex.madness");
	// init OpenGL using GLUT

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutMouseFunc(onMouseClick);
	glutMotionFunc(onMouseMotion);
	// register callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
